///////////////////////////////////////////////////////////////////////////////
//  untAdderProxy:                                                           //
//  Proxy object to wrap a method call inside some SOAP XML, send it with    //
//  HTTP, extract a returned result from the HTTP response and return the    //
//  result as a string.                                                      //
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//  Written for the January/February 2001 issue of                           //
//  "uk-bug developer's magazine"                                            //
//  by Rob Bracken                                                           //
//  Bracken Software Limited                                                 //
//  01275 843215                                                             //
///////////////////////////////////////////////////////////////////////////////

unit UntAdderProxy;

interface

uses
  Classes,
  NMHTTP,
  XDOM;

type

  TAdder = class(TObject)
  private

    FXML: TStrings;
    FHTTPHeader: TStringList;
    FhttpSOAP: TNMHTTP;
    FXMLParser: TXmlToDomParser;
    FDomImpl: TDomImplementation;
    FDomDoc: TDOMDocument;

    function BuildXML(Int1, Int2: integer): TStringList;
    function GetResult(Response, MethodName: string): string;
    procedure AboutToSend(Sender: TObject);
    procedure HTTPFailure(Cmd: CmdType);
    procedure HTTPSuccess(Cmd: CmdType);

  protected

  public

    constructor Create;
    destructor Destroy; override;
    function Add(Int1, Int2: integer): String;

  published

  end;

implementation

{ TAdder }

uses
  untClientForm,                        // Included so we can display the HTTP
                                        // headers, request & response.
  Dialogs,
  SysUtils;

procedure TAdder.AboutToSend(Sender: TObject);
{ Add a SOAPAction parameter to the POST header }
begin
  if Sender is TNMHTTP then
  begin
    TNMHTTP(Sender).SendHeader.Add('SOAPAction:"SOAPAdder"');
    FHTTPHeader.Add('--- Header: ---');
    FHTTPHeader.AddStrings(TNMHTTP(Sender).SendHeader);
//     TNMHTTP(Sender).SendHeader.Values['SOAPAction'] := 'SOAPAdder';
  end;
end;

function TAdder.Add(Int1, Int2: integer): String;
{ Package the call to the Add method in a SOAP request and use HTTP to send it
  to a CGI program which will unpackage it and call the method on the real
  object }
begin

  // Build the XML SOAP request
  FXML := BuildXML(Int1, Int2);
  try
    // Send it via HTTP
    FhttpSOAP.OutputFileMode := False;
    FhttpSOAP.InputFileMode := False;
    FhttpSOAP.Post('http://localhost/scripts/SOAPServer.exe', FXML.Text);
    // Display the HTTP header
    frmClient.mmoRequest.Lines.AddStrings(FHTTPHeader);
    // Display the SOAP request
    frmClient.mmoRequest.Lines.Add('--- XML request: ---');
    frmClient.mmoRequest.Lines.Add(FXML.Text);
    // Extract the return value from the HTTP response
    Result := GetResult(FhttpSOAP.Body, 'Add');
    // Display the returned XML
    frmClient.mmoResponse.Lines.Add('--- HTTP Header: ---');
    frmClient.mmoResponse.Lines.Add(FhttpSOAP.Header);
    frmClient.mmoResponse.Lines.Add('--- XML Response: ---');
    frmClient.mmoResponse.Lines.Add(FhttpSOAP.Body);
  finally
    FXML.Free;
  end;

end;

function TAdder.BuildXML(Int1, Int2: integer): TStringList;
{ Build the XML for the SOAP request }
begin
  // The XML doesn't have to be on separate lines
  Result := TStringList.Create;
  Result.Add('<SOAP-ENV:Envelope');
  Result.Add(  'xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/"');
  Result.Add(  'SOAP-ENV:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">');
  Result.Add(   '<SOAP-ENV:Body>');
  Result.Add(       '<m:Add xmlns:m="SOAPAdder">');
  Result.Add(           '<Int1>' + IntToStr(Int1) + '</Int1>');
  Result.Add(           '<Int2>' + IntToStr(Int2) + '</Int2>');
  Result.Add(       '</m:Add>');
  Result.Add(   '</SOAP-ENV:Body>');
  Result.Add('</SOAP-ENV:Envelope>');
end;

constructor TAdder.Create;
begin

  // Create the XML parsing components
  FXMLParser := TXmlToDomParser.Create(nil);
  FDomImpl := TDomImplementation.Create(nil);
  FXMLParser.DOMImpl := FDomImpl;

  // Create an HTTP component to send an HTTP POST
  FhttpSOAP := TNMHTTP.Create(nil);
  FHTTPHeader := TStringList.Create;        // Somewhere to save the HTTP Header

  // Attach a few event handlers
  FhttpSOAP.OnAboutToSend := AboutToSend;
  FhttpSOAP.OnFailure := HTTPFailure;
  FhttpSOAP.OnSuccess := HTTPSuccess;

end;

destructor TAdder.Destroy;
begin
  // Free the HTTP component
  FhttpSOAP.Free;
  FHTTPHeader.Free;
  // Free the XML parsing components
  FXMLParser.Free;
  FDomImpl.Free;
  inherited;

end;

function TAdder.GetResult(Response, MethodName: string): string;
{ Parse the XML response and extract a method result
  Response   - the XML response from the SOAP method call
  MethodName - the name of the method whose result we want }

  function GetNextXMLElement(XMLNode: TDOMNode): TDOMNode;
  { Find the _next_ XML element node at this level (NodeType = ntElement_Node) }
  begin
    Result := XMLNode.NextSibling;
    while (Result.NodeType <> ntElement_Node) and (Result <> nil) do
      Result := Result.NextSibling;
  end;

  function FindNode(XMLNode: TDOMNode; sValueToFind: string): TDOMNode;
  { Find the first XML element at this level that has the supplied value,
    starting with the supplied node. (An XML element is a node with a NodeType
    of ntElement_Node)
    Note that we actually look at the NodeName property, rather than the
    NodeValue. }
  begin

    if XMLNode.NodeType <> ntElement_Node then
    begin
      Result := GetNextXMLElement(XMLNode);
    end
    else
    begin
      Result := XMLNode;
    end;
    // Find the first child that has a matching value
    while (Result.NodeName <> sValueToFind) and (Result <> nil) do
      Result := GetNextXMLElement(XMLNode);
  end;

  function GetMethodName(SOAPMethodNode: TDOMNode): string;
  { Extract the method name from the node name. The node name is in the form
    "ns:method" where ns = the XML namespace and method = the method name,
    so we need to extract everything after the colon. }
  var
    nColonPos: integer;
  begin
    nColonPos := Pos(':',SOAPMethodNode.NodeName);
    Result := Copy(SOAPMethodNode.NodeName,nColonPos+1,
                   length(SOAPMethodNode.NodeName)-nColonPos);
  end;

  function FindMethodNode(XMLNode: TDOMNode; sMethodName: string): TDOMNode;
  { Find the first XML element at this level that has the supplied method name,
    starting with the supplied node. (An XML element is a node with a NodeType
    of ntElement_Node)
    Note that we actually look at the NodeName property, rather than the
    NodeValue. }
  begin

    if XMLNode.NodeType <> ntElement_Node then
    begin
      Result := GetNextXMLElement(XMLNode);
    end
    else
    begin
      Result := XMLNode;
    end;
    // Find the first child that has a matching value
    while (GetMethodName(Result) <> sMethodName) and (Result <> nil) do
      Result := GetNextXMLElement(XMLNode);
  end;

var
  SOAPEnvelopeNode, SOAPBodyNode, SOAPMethodNode, tmpNode: TDomNode;
const
  SyntaxStuff = [WideChar(#9), WideChar(#10), WideChar(#13)];
begin

  Result := '';

  // Parse the XML and load it into a DOM document tree
  FDomDoc := FXMLParser.stringToDom(Response);

  // Find the SOAP envelope node (will be a child of the DOM document node)
  SOAPEnvelopeNode := FindNode(FDomDoc.FirstChild, 'SOAP-ENV:Envelope');

  // Find the SOAP Body node (will be a child of the envelope node)
  SOAPBodyNode := FindNode(SOAPEnvelopeNode.FirstChild, 'SOAP-ENV:Body');

  // Method nodes will be children of the SOAP Body node.
  // For simplicity, we've assumed that all the method calls are for us, so we
  // don't check the namespace. To check the namespace, find the "xmlns"
  // attribute of the method node and compare its value.
  // We've also assumed that there will only be one result for a particular
  // method.
  SOAPMethodNode := FindMethodNode(SOAPBodyNode.FirstChild, MethodName);
  if Assigned(SOAPMethodNode) then
  begin
    // Get the first child node that is an XML element. This will be the result
    // node.
    tmpNode := SOAPMethodNode.FirstChild;
    if tmpNode.NodeType <> ntElement_Node then
      tmpNode := GetNextXMLElement(SOAPMethodNode.FirstChild);
    // The result value will be the first child of the result node that is not
    // a padding node (such as a CR/LF).
    tmpNode := tmpNode.FirstChild;
    while (tmpNode.NodeType <> ntText_Node) or (tmpNode.NodeValue[1] in SyntaxStuff) do
      tmpNode := tmpNode.NextSibling;
    Result := tmpNode.NodeValue;
  end;

end;

procedure TAdder.HTTPFailure(Cmd: CmdType);
{ Things to do when the HTTP request fails }
begin

  // Process failures
  ShowMessage('SOAP method call failed');

end;

procedure TAdder.HTTPSuccess(Cmd: CmdType);
{ Things to do when the HTTP request succeeds }
begin

  // When we've got a successful response, process it
  // Ignore everything except responses from POST requests
  case Cmd of
    cmdPost: ShowMessage('SOAP method call was successful');
  end;

end;

end.
